@@ -17,7 +17,7 @@
  *
  *   You should have received a copy of the GNU General Public License a
  *   with this program; if not, write to the Free Software Foundation, Inc.,
- *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
+ *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */   
 
 #include <config.h>
@@ -42,6 +42,7 @@
 #include <sys/sendfile.h>
 #include <sys/xattr.h>
 #include <sys/timeb.h>
+#include <sys/mman.h>
 #include <bits/wordsize.h>
 #include <utime.h>
 #include <stdarg.h>
@@ -61,7 +62,18 @@
 sfun _pure_syscall=syscall;
 sfun _pure_native_syscall=syscall;
 
-int _pure_debug_printf(const char *format, ...)
+static int _pageshift()
+{
+	static int ps=0;
+	if (ps == 0) {
+		long pagesize=pagesize = sysconf(_SC_PAGESIZE);
+		for (ps = -1;pagesize > 0; ps++, pagesize >>= 1)
+			;
+	}
+	return ps;
+}
+
+long _pure_debug_printf(const char *format, ...)
 {
 	char *s;
 	int rv;
@@ -279,7 +291,7 @@ int access(const char* pathname,int mode){
 	return _pure_syscall(__NR_access,pathname,mode);
 }
 
-int readlink(const char* pathname,char* buf, size_t bufsize){
+ssize_t readlink(const char* pathname,char* buf, size_t bufsize){
 	return _pure_syscall(__NR_readlink,pathname,buf,bufsize);
 }
 
@@ -1007,17 +1019,58 @@ static void statfs2vfs(struct statfs *sfs,struct statvfs *vsfs)
 int statvfs(const char *path, struct statvfs *buf){
 	struct statfs sfs;
 	int rv=_pure_syscall(__NR_statfs,path,&sfs);
-	statfs2vfs(&sfs,buf);
-	return 0;
+	if (rv >= 0) statfs2vfs(&sfs,buf);
+	return rv;
 }
 
 int fstatvfs(int fd, struct statvfs *buf){
 	struct statfs sfs;
 	int rv=_pure_syscall(__NR_fstatfs,fd,&sfs);
-	statfs2vfs(&sfs,buf);
-	return 0;
+	if (rv >= 0) statfs2vfs(&sfs,buf);
+	return rv;
 }
 
+void *mmap(void  *start, size_t length, int prot, int flags, int fd,
+		       off_t offset)
+{
+#if __WORDSIZE != 64
+		return (void *) _pure_syscall(__NR_mmap2,start,length,prot,flags,fd,offset>> _pageshift());
+#else
+		return (void *) _pure_syscall(__NR_mmap,start,length,prot,flags,fd,offset);
+#endif
+}
+
+#if __WORDSIZE != 64
+void *mmap2(void  *start, size_t length, int prot, int flags, int fd,
+		       off_t pgoffset)
+{
+		return (void *) _pure_syscall(__NR_mmap2,start,length,prot,flags,fd,pgoffset);
+}
+#endif
+
+int munmap(void *start, size_t length)
+{
+	return _pure_syscall(__NR_munmap,start,length);
+}
+
+#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 5
+void *mremap(void  *old_address,  size_t old_size , size_t new_size,
+		       int flags, ...)
+{
+	va_list ap;
+	va_start (ap, flags);
+	void *newaddr = (flags & MREMAP_FIXED) ? va_arg (ap, void *) : NULL;
+	va_end (ap);
+	return (void *) _pure_syscall(__NR_mremap,old_address,old_size,new_size,flags,newaddr);
+}
+#else
+void *mremap(void  *old_address,  size_t old_size , size_t new_size,
+		       int flags)
+{
+	return (void *) _pure_syscall(__NR_mremap,old_address,old_size,new_size,flags);
+}
+#endif
+
 int ftime(struct timeb *tp){
 	struct timeval tv;
 	struct timezone tz;
